@@ -45,9 +45,10 @@
 #include <rpc/xdr.h>
 #include <rpc/auth.h>
 #include <rpc/auth_gss.h>
+#include <rpc/rpcsec_gss.h>
 #include <rpc/clnt.h>
 #include <netinet/in.h>
-#include <gssapi/gssapi.h>
+#include <reentrant.h>
 
 #include "debug.h"
 
@@ -60,6 +61,9 @@ static void	authgss_destroy_context(AUTH *);
 static bool_t	authgss_wrap(AUTH *, XDR *, xdrproc_t, caddr_t);
 static bool_t	authgss_unwrap(AUTH *, XDR *, xdrproc_t, caddr_t);
 
+void rpc_gss_set_error(int);
+void rpc_gss_clear_error(void);
+bool_t rpc_gss_oid_to_mech(rpc_gss_OID, char **);
 
 /*
  * from mit-krb5-1.2.1 mechglue/mglueP.h:
@@ -131,6 +135,8 @@ char *p;
 	fprintf(stderr, "     cred: %p\n", ptr->cred);
 }
 
+extern pthread_mutex_t		 auth_ref_lock;
+
 struct rpc_gss_data {
 	bool_t			 established;	/* context established */
 	gss_buffer_desc		 gc_wire_verf;	/* save GSS_S_COMPLETE NULL RPC verfier
@@ -141,12 +147,37 @@ struct rpc_gss_data {
 	gss_ctx_id_t		 ctx;		/* context id */
 	struct rpc_gss_cred	 gc;		/* client credentials */
 	u_int			 win;		/* sequence window */
+	int			 time_req;	/* init_sec_context time_req */
+	gss_channel_bindings_t	 icb;		/* input channel bindings */
+	int			 refcnt;	/* reference count gss AUTHs */
 };
 
 #define	AUTH_PRIVATE(auth)	((struct rpc_gss_data *)auth->ah_private)
 
 static struct timeval AUTH_TIMEOUT = { 25, 0 };
 
+static void
+authgss_auth_get(AUTH *auth)
+{
+	struct rpc_gss_data	*gd = AUTH_PRIVATE(auth);
+
+	mutex_lock(&auth_ref_lock);
+	++gd->refcnt;
+	mutex_unlock(&auth_ref_lock);
+}
+
+static int
+authgss_auth_put(AUTH *auth)
+{
+	int refcnt;
+	struct rpc_gss_data	*gd = AUTH_PRIVATE(auth);
+
+	mutex_lock(&auth_ref_lock);
+	refcnt = --gd->refcnt;
+	mutex_unlock(&auth_ref_lock);
+	return refcnt;
+}
+
 AUTH *
 authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 {
@@ -200,7 +231,7 @@ authgss_create(CLIENT *clnt, gss_name_t name, struct rpc_gss_sec *sec)
 	if (!authgss_refresh(auth, NULL))
 		auth = NULL;
 	else
-		auth_get(auth); /* Reference for caller */
+		authgss_auth_get(auth); /* Reference for caller */
 
 	clnt->cl_auth = save_auth;
 
@@ -419,12 +450,15 @@ authgss_validate(AUTH *auth, struct opaque_auth *verf)
 }
 
 static bool_t
-authgss_refresh(AUTH *auth, void *dummy)
+_rpc_gss_refresh(AUTH *auth, rpc_gss_options_ret_t *options_ret)
 {
 	struct rpc_gss_data	*gd;
 	struct rpc_gss_init_res	 gr;
 	gss_buffer_desc		*recv_tokenp, send_token;
-	OM_uint32		 maj_stat, min_stat, call_stat, ret_flags;
+	OM_uint32		 maj_stat, min_stat, call_stat, ret_flags,
+				 time_ret;
+	gss_OID			 actual_mech_type;
+	char			*mechanism;
 
 	gss_log_debug("in authgss_refresh()");
 
@@ -452,13 +486,13 @@ authgss_refresh(AUTH *auth, void *dummy)
 						gd->name,
 						gd->sec.mech,
 						gd->sec.req_flags,
-						0,		/* time req */
-						NULL,		/* channel */
+						gd->time_req,
+						gd->icb,
 						recv_tokenp,
-						NULL,		/* used mech */
+						&actual_mech_type,
 						&send_token,
 						&ret_flags,
-						NULL);		/* time rec */
+						&time_ret);
 
 		if (recv_tokenp != GSS_C_NO_BUFFER) {
 			gss_release_buffer(&min_stat, &gr.gr_token);
@@ -466,8 +500,9 @@ authgss_refresh(AUTH *auth, void *dummy)
 		}
 		if (maj_stat != GSS_S_COMPLETE &&
 		    maj_stat != GSS_S_CONTINUE_NEEDED) {
-			gss_log_status("authgss_refresh: gss_init_sec_context", 
-				maj_stat, min_stat);
+			gss_log_status("gss_init_sec_context", maj_stat, min_stat);
+			options_ret->major_status = maj_stat;
+			options_ret->minor_status = min_stat;
 			break;
 		}
 		if (send_token.length != 0) {
@@ -488,8 +523,11 @@ authgss_refresh(AUTH *auth, void *dummy)
 
 			if (call_stat != RPC_SUCCESS ||
 			    (gr.gr_major != GSS_S_COMPLETE &&
-			     gr.gr_major != GSS_S_CONTINUE_NEEDED))
+			     gr.gr_major != GSS_S_CONTINUE_NEEDED)) {
+				options_ret->major_status = gr.gr_major;
+				options_ret->minor_status = gr.gr_minor;
 				return FALSE;
+			}
 
 			if (gr.gr_ctx.length != 0) {
 				if (gd->gc.gc_ctx.value)
@@ -530,8 +568,25 @@ authgss_refresh(AUTH *auth, void *dummy)
 					gd->established = FALSE;
 					authgss_destroy_context(auth);
 				}
+				rpc_gss_set_error(EPERM);
+				options_ret->major_status = maj_stat;
+				options_ret->minor_status = min_stat;
 				return (FALSE);
 			}
+
+			options_ret->major_status = GSS_S_COMPLETE;
+			options_ret->minor_status = 0;
+			options_ret->rpcsec_version = gd->gc.gc_v;
+			options_ret->ret_flags = ret_flags;
+			options_ret->time_ret = time_ret;
+			options_ret->gss_context = gd->ctx;
+			options_ret->actual_mechanism[0] = '\0';
+			if (rpc_gss_oid_to_mech(actual_mech_type, &mechanism)) {
+				strncpy(options_ret->actual_mechanism,
+					mechanism,
+					sizeof(options_ret->actual_mechanism));
+			}
+
 			gd->established = TRUE;
 			gd->gc.gc_proc = RPCSEC_GSS_DATA;
 			gd->gc.gc_seq = 0;
@@ -547,12 +602,22 @@ authgss_refresh(AUTH *auth, void *dummy)
 		authgss_destroy(auth);
 		auth = NULL;
 		rpc_createerr.cf_stat = RPC_AUTHERROR;
+		rpc_gss_set_error(EPERM);
 
 		return (FALSE);
 	}
 	return (TRUE);
 }
 
+static bool_t
+authgss_refresh(AUTH *auth, void *dummy)
+{
+	rpc_gss_options_ret_t ret;
+
+	memset(&ret, 0, sizeof(ret));
+	return _rpc_gss_refresh(auth, &ret);
+}
+
 bool_t
 authgss_service(AUTH *auth, int svc)
 {
@@ -617,21 +682,24 @@ authgss_destroy_context(AUTH *auth)
 static void
 authgss_destroy(AUTH *auth)
 {
-	struct rpc_gss_data	*gd;
-	OM_uint32		 min_stat;
+	if(authgss_auth_put(auth) == 0)
+	{
+		struct rpc_gss_data	*gd;
+		OM_uint32		 min_stat;
 
-	gss_log_debug("in authgss_destroy()");
+		gss_log_debug("in authgss_destroy()");
 
-	gd = AUTH_PRIVATE(auth);
+		gd = AUTH_PRIVATE(auth);
 
-	authgss_destroy_context(auth);
+		authgss_destroy_context(auth);
 
-	LIBTIRPC_DEBUG(3, ("authgss_destroy: freeing name %p", gd->name));
-	if (gd->name != GSS_C_NO_NAME)
-		gss_release_name(&min_stat, &gd->name);
+		LIBTIRPC_DEBUG(3, ("authgss_destroy: freeing name %p", gd->name));
+		if (gd->name != GSS_C_NO_NAME)
+			gss_release_name(&min_stat, &gd->name);
 
-	free(gd);
-	free(auth);
+		free(gd);
+		free(auth);
+	}
 }
 
 static bool_t
@@ -667,3 +735,241 @@ authgss_unwrap(AUTH *auth, XDR *xdrs, xdrproc_t xdr_func, caddr_t xdr_ptr)
 				 gd->ctx, gd->sec.qop,
 				 gd->sec.svc, gd->gc.gc_seq));
 }
+
+static AUTH *
+_rpc_gss_seccreate_error(int system_error)
+{
+	rpc_gss_set_error(system_error);
+	return NULL;
+}
+
+/*
+ * External API: Create a GSS security context for this "clnt"
+ *
+ * clnt: a valid RPC CLIENT structure
+ * principal: NUL-terminated C string containing GSS principal
+ * mechanism: NUL-terminated C string containing GSS mechanism name
+ * service: GSS security value to use
+ * qop: NUL-terminated C string containing QOP name
+ * req: pointer to additional request parameters, or NULL
+ * ret: pointer to additional results, or NULL
+ *
+ * Returns pointer to a filled in RPC AUTH structure or NULL.
+ * Caller must free returned structure with AUTH_DESTROY.
+ */
+AUTH *
+rpc_gss_seccreate(CLIENT *clnt, char *principal, char *mechanism,
+		rpc_gss_service_t service, char *qop,
+		rpc_gss_options_req_t *req, rpc_gss_options_ret_t *ret)
+{
+	rpc_gss_options_ret_t options_ret;
+	OM_uint32 maj_stat, min_stat;
+	struct rpc_gss_sec sec = {
+		.req_flags	= GSS_C_MUTUAL_FLAG,
+		.cred		= GSS_C_NO_CREDENTIAL,
+	};
+	struct rpc_gss_data *gd;
+	AUTH *auth, *save_auth;
+	gss_buffer_desc sname;
+
+	if (clnt == NULL || principal == NULL)
+		return _rpc_gss_seccreate_error(EINVAL);
+
+	if (rpc_gss_mech_to_oid(mechanism, &sec.mech) == FALSE)
+		return NULL;
+
+	sec.qop = GSS_C_QOP_DEFAULT;
+	if (qop != NULL) {
+		u_int qop_num;
+		if (rpc_gss_qop_to_num(qop, mechanism, &qop_num) == FALSE)
+			return NULL;
+		sec.qop = qop_num;
+	}
+
+	switch (service) {
+	case rpcsec_gss_svc_none:
+		sec.svc = RPCSEC_GSS_SVC_NONE;
+		break;
+	case rpcsec_gss_svc_integrity:
+		sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	case rpcsec_gss_svc_privacy:
+		sec.svc = RPCSEC_GSS_SVC_PRIVACY;
+		break;
+	case rpcsec_gss_svc_default:
+		sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	default:
+		return _rpc_gss_seccreate_error(ENOENT);
+	}
+
+	if (ret == NULL)
+		ret = &options_ret;
+	memset(ret, 0, sizeof(*ret));
+
+	auth = calloc(1, sizeof(*auth));
+	gd = calloc(1, sizeof(*gd));
+	if (auth == NULL || gd == NULL) {
+		free(gd);
+		free(auth);
+		return _rpc_gss_seccreate_error(ENOMEM);
+	}
+
+	sname.value = principal;
+	sname.length = strlen(principal);
+	maj_stat = gss_import_name(&min_stat, &sname,
+					(gss_OID)GSS_C_NT_HOSTBASED_SERVICE,
+					&gd->name);
+	if (maj_stat != GSS_S_COMPLETE) {
+		ret->major_status = maj_stat;
+		ret->minor_status = min_stat;
+		free(gd);
+		free(auth);
+		return _rpc_gss_seccreate_error(ENOMEM);
+	}
+
+	gd->clnt = clnt;
+	gd->ctx = GSS_C_NO_CONTEXT;
+	gd->sec = sec;
+
+	if (req) {
+		sec.req_flags = req->req_flags;
+		gd->time_req = req->time_req;
+		sec.cred = req->my_cred;
+		gd->icb = req->input_channel_bindings;
+	}
+
+	gd->gc.gc_v = RPCSEC_GSS_VERSION;
+	gd->gc.gc_proc = RPCSEC_GSS_INIT;
+	gd->gc.gc_svc = sec.svc;
+
+	auth->ah_ops = &authgss_ops;
+	auth->ah_private = (caddr_t)gd;
+
+	save_auth = clnt->cl_auth;
+	clnt->cl_auth = auth;
+
+	if (_rpc_gss_refresh(auth, ret) == FALSE) {
+		auth = NULL;
+	} else {
+		rpc_gss_clear_error();
+		authgss_auth_get(auth);
+	}
+
+	clnt->cl_auth = save_auth;
+
+	return auth;
+}
+
+/*
+ * External API: Modify an AUTH's service and QOP settings
+ *
+ * auth: a valid RPC AUTH structure
+ * service: GSS security value to use
+ * qop: NUL-terminated C string containing QOP name
+ *
+ * Returns TRUE if successful, otherwise FALSE is returned.
+ */
+bool_t
+rpc_gss_set_defaults(AUTH *auth, rpc_gss_service_t service, char *qop)
+{
+	struct rpc_gss_data *gd;
+	char *mechanism;
+	u_int qop_num;
+
+	if (auth == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return FALSE;
+	}
+
+	gd = AUTH_PRIVATE(auth);
+	if (gd == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return FALSE;
+	}
+
+	if (rpc_gss_oid_to_mech(gd->sec.mech, &mechanism) == FALSE)
+		return FALSE;
+
+	qop_num = GSS_C_QOP_DEFAULT;
+	if (qop != NULL) {
+		if (rpc_gss_qop_to_num(qop, mechanism, &qop_num) == FALSE)
+			return FALSE;
+	}
+
+	switch (service) {
+	case rpcsec_gss_svc_none:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_NONE;
+		break;
+	case rpcsec_gss_svc_integrity:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	case rpcsec_gss_svc_privacy:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_PRIVACY;
+		break;
+	case rpcsec_gss_svc_default:
+		gd->gc.gc_svc = gd->sec.svc = RPCSEC_GSS_SVC_INTEGRITY;
+		break;
+	default:
+		rpc_gss_set_error(ENOENT);
+		return FALSE;
+	}
+
+	gd->sec.qop = (gss_qop_t)qop_num;
+	rpc_gss_clear_error();
+	return TRUE;
+}
+
+/*
+ * External API: Return maximum data size for a security mechanism and transport
+ *
+ * auth: a valid RPC AUTH structure
+ * maxlen: transport's maximum data size, in bytes
+ *
+ * Returns maximum data size given transformations done by current
+ * security setting of "auth", in bytes, or zero if that value
+ * cannot be determined.
+ */
+int
+rpc_gss_max_data_length(AUTH *auth, int maxlen)
+{
+	OM_uint32 max_input_size, maj_stat, min_stat;
+	struct rpc_gss_data *gd;
+	int conf_req_flag;
+	int result;
+
+	if (auth == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return 0;
+	}
+
+	gd = AUTH_PRIVATE(auth);
+	if (gd == NULL) {
+		rpc_gss_set_error(EINVAL);
+		return 0;
+	}
+
+	switch (gd->sec.svc) {
+	case RPCSEC_GSS_SVC_NONE:
+		rpc_gss_clear_error();
+		return maxlen;
+	case RPCSEC_GSS_SVC_INTEGRITY:
+		conf_req_flag = 0;
+		break;
+	case RPCSEC_GSS_SVC_PRIVACY:
+		conf_req_flag = 1;
+		break;
+	default:
+		rpc_gss_set_error(ENOENT);
+		return 0;
+	}
+
+	result = 0;
+	maj_stat = gss_wrap_size_limit(&min_stat, gd->ctx, conf_req_flag,
+					gd->sec.qop, maxlen, &max_input_size);
+	if (maj_stat == GSS_S_COMPLETE)
+		if ((int)max_input_size > 0)
+			result = (int)max_input_size;
+	rpc_gss_clear_error();
+	return result;
+}
