@@ -48,7 +48,7 @@
 
 #include <reentrant.h>
 #include <sys/types.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <sys/syslog.h>
 #include <sys/un.h>
 #include <sys/uio.h>
