@@ -27,7 +27,6 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 
-#include <sys/cdefs.h>
 
 /*
  * svc_vc.c, Server side for Connection Oriented based RPC. 
@@ -41,7 +40,7 @@
 #include <sys/socket.h>
 #include <sys/types.h>
 #include <sys/param.h>
-#include <sys/poll.h>
+#include <poll.h>
 #include <sys/un.h>
 #include <sys/time.h>
 #include <sys/uio.h>
@@ -147,7 +146,8 @@ svc_vc_create(fd, sendsize, recvsize)
 	u_int sendsize;
 	u_int recvsize;
 {
-	SVCXPRT *xprt;
+	SVCXPRT *xprt = NULL;
+	SVCXPRT_EXT *ext = NULL;
 	struct cf_rendezvous *r = NULL;
 	struct __rpc_sockinfo si;
 	struct sockaddr_storage sslocal;
@@ -158,8 +158,10 @@ svc_vc_create(fd, sendsize, recvsize)
 		warnx("svc_vc_create: out of memory");
 		goto cleanup_svc_vc_create;
 	}
-	if (!__rpc_fd2sockinfo(fd, &si))
-		return NULL;
+	if (!__rpc_fd2sockinfo(fd, &si)) {
+		warnx("svc_vc_create: __rpc_fd2sockinfo failed");
+		goto cleanup_svc_vc_create;
+	}
 	r->sendsize = __rpc_get_t_size(si.si_af, si.si_proto, (int)sendsize);
 	r->recvsize = __rpc_get_t_size(si.si_af, si.si_proto, (int)recvsize);
 	r->maxrec = __svc_maxrec;
@@ -168,11 +170,16 @@ svc_vc_create(fd, sendsize, recvsize)
 		warnx("svc_vc_create: out of memory");
 		goto cleanup_svc_vc_create;
 	}
+	ext = mem_alloc(sizeof (*ext));
+	if (ext == NULL) {
+		warnx("svc_vc_create: out of memory");
+		goto cleanup_svc_vc_create;
+	}
+	memset(ext, 0, sizeof (*ext));
 	xprt->xp_tp = NULL;
 	xprt->xp_p1 = r;
 	xprt->xp_p2 = NULL;
-	xprt->xp_p3 = NULL;
-	xprt->xp_auth = NULL;
+	xprt->xp_p3 = ext;
 	xprt->xp_verf = _null_auth;
 	svc_vc_rendezvous_ops(xprt);
 	xprt->xp_port = (u_short)-1;	/* It is the rendezvouser */
@@ -193,6 +200,10 @@ svc_vc_create(fd, sendsize, recvsize)
 cleanup_svc_vc_create:
 	if (r != NULL)
 		mem_free(r, sizeof(*r));
+	if (xprt != NULL)
+		mem_free(xprt, sizeof(*xprt));
+	if (ext != NULL)
+		mem_free(ext, sizeof(*ext));
 	return (NULL);
 }
 
@@ -255,6 +266,7 @@ makefd_xprt(fd, sendsize, recvsize)
 	u_int recvsize;
 {
 	SVCXPRT *xprt;
+	SVCXPRT_EXT *ext;
 	struct cf_conn *cd;
 	const char *netid;
 	struct __rpc_sockinfo si;
@@ -273,9 +285,18 @@ makefd_xprt(fd, sendsize, recvsize)
 		goto done;
 	}
 	memset(xprt, 0, sizeof *xprt);
+	ext = mem_alloc(sizeof (*ext));
+	if (ext == NULL) {
+		warnx("svc_vc: makefd_xprt: out of memory");
+		mem_free(xprt, sizeof(SVCXPRT));
+		xprt = NULL;
+		goto done;
+	}
+	memset(ext, 0, sizeof (*ext));
 	cd = mem_alloc(sizeof(struct cf_conn));
 	if (cd == NULL) {
-		warnx("svc_tcp: makefd_xprt: out of memory");
+		warnx("svc_vc: makefd_xprt: out of memory");
+		mem_free(ext, sizeof(*ext));
 		mem_free(xprt, sizeof(SVCXPRT));
 		xprt = NULL;
 		goto done;
@@ -284,7 +305,7 @@ makefd_xprt(fd, sendsize, recvsize)
 	xdrrec_create(&(cd->xdrs), sendsize, recvsize,
 	    xprt, read_vc, write_vc);
 	xprt->xp_p1 = cd;
-	xprt->xp_auth = NULL;
+	xprt->xp_p3 = ext;
 	xprt->xp_verf.oa_base = cd->verf_body;
 	svc_vc_ops(xprt);  /* truely deals with calls */
 	xprt->xp_port = 0;  /* this is a connection, not a rendezvouser */
@@ -310,7 +331,6 @@ rendezvous_request(xprt, msg)
 	socklen_t len;
 	struct __rpc_sockinfo si;
 	SVCXPRT *newxprt;
-	fd_set cleanfds;
 
 	assert(xprt != NULL);
 	assert(msg != NULL);
@@ -322,13 +342,16 @@ again:
 	    &len)) < 0) {
 		if (errno == EINTR)
 			goto again;
-		/*
-		 * Clean out the most idle file descriptor when we're
-		 * running out.
-		 */
+
 		if (errno == EMFILE || errno == ENFILE) {
-			cleanfds = svc_fdset;
-			__svc_clean_idle(&cleanfds, 0, FALSE);
+		  /* If there are no file descriptors available, then accept will fail.
+		     We want to delay here so the connection request can be dequeued;
+		     otherwise we can bounce between polling and accepting, never
+		     giving the request a chance to dequeue and eating an enormous
+		     amount of cpu time in svc_run if we're polling on many file
+		     descriptors.  */
+		        struct timespec ts = { .tv_sec = 0, .tv_nsec = 50000000 };
+                        nanosleep (&ts, NULL);
 			goto again;
 		}
 		return (FALSE);
@@ -338,6 +361,8 @@ again:
 	 */
 
 	newxprt = makefd_xprt(sock, r->sendsize, r->recvsize);
+	if (!newxprt)
+		return (FALSE);
 
 	if (!__rpc_set_netbuf(&newxprt->xp_rtaddr, &addr, len))
 		return (FALSE);
@@ -393,10 +418,17 @@ svc_vc_destroy(xprt)
 	__svc_vc_dodestroy(xprt);
 }
 
+static bool_t
+__svc_rendezvous_socket(xprt)
+	SVCXPRT *xprt;
+{
+	return (xprt->xp_ops->xp_recv == rendezvous_request);
+}
 static void
 __svc_vc_dodestroy(xprt)
 	SVCXPRT *xprt;
 {
+	SVCXPRT_EXT *ext = SVCEXT(xprt);
 	struct cf_conn *cd;
 	struct cf_rendezvous *r;
 
@@ -404,7 +436,7 @@ __svc_vc_dodestroy(xprt)
 
 	if (xprt->xp_fd != RPC_ANYFD)
 		(void)close(xprt->xp_fd);
-	if (xprt->xp_port != 0) {
+	if (__svc_rendezvous_socket(xprt)) {
 		/* a rendezvouser socket */
 		r = (struct cf_rendezvous *)xprt->xp_p1;
 		mem_free(r, sizeof (struct cf_rendezvous));
@@ -414,10 +446,8 @@ __svc_vc_dodestroy(xprt)
 		XDR_DESTROY(&(cd->xdrs));
 		mem_free(cd, sizeof(struct cf_conn));
 	}
-	if (xprt->xp_auth != NULL) {
-		SVCAUTH_DESTROY(xprt->xp_auth);
-		xprt->xp_auth = NULL;
-	}
+	if (ext)
+		mem_free(ext, sizeof (*ext));
 	if (xprt->xp_rtaddr.buf)
 		mem_free(xprt->xp_rtaddr.buf, xprt->xp_rtaddr.maxlen);
 	if (xprt->xp_ltaddr.buf)
@@ -559,20 +589,19 @@ write_vc(xprtp, buf, len)
 				cd->strm_stat = XPRT_DIED;
 				return (-1);
 			}
-			if (cd->nonblock && i != cnt) {
-				/*
-				 * For non-blocking connections, do not
-				 * take more than 2 seconds writing the
-				 * data out.
-				 *
-				 * XXX 2 is an arbitrary amount.
-				 */
-				gettimeofday(&tv1, NULL);
-				if (tv1.tv_sec - tv0.tv_sec >= 2) {
-					cd->strm_stat = XPRT_DIED;
-					return (-1);
-				}
+			/*
+			 * For non-blocking connections, do not
+			 * take more than 2 seconds writing the
+			 * data out.
+			 *
+			 * XXX 2 is an arbitrary amount.
+			 */
+			gettimeofday(&tv1, NULL);
+			if (tv1.tv_sec - tv0.tv_sec >= 2) {
+				cd->strm_stat = XPRT_DIED;
+				return (-1);
 			}
+			i = 0; /* Don't change buf and cnt */
 		}
 	}
 
@@ -639,9 +668,9 @@ svc_vc_getargs(xprt, xdr_args, args_ptr)
 	assert(xprt != NULL);
 	/* args_ptr may be NULL */
 
-	if (! SVCAUTH_UNWRAP(xprt->xp_auth,
-			     &(((struct cf_conn *)(xprt->xp_p1))->xdrs),
-			     xdr_args, args_ptr)) {
+	if (!SVCAUTH_UNWRAP(&SVC_XP_AUTH(xprt),
+			    &(((struct cf_conn *)(xprt->xp_p1))->xdrs),
+			    xdr_args, args_ptr)) {
 		return FALSE;  
 	}
 	return TRUE;
@@ -698,8 +727,9 @@ svc_vc_reply(xprt, msg)
 	msg->rm_xid = cd->x_id;
 	rstat = FALSE;
 	if (xdr_replymsg(xdrs, msg) &&
-	    (!has_args || (xprt->xp_auth &&
-	     SVCAUTH_WRAP(xprt->xp_auth, xdrs, xdr_results, xdr_location)))) {
+	    (!has_args ||
+	     SVCAUTH_WRAP(&SVC_XP_AUTH(xprt),
+			  xdrs, xdr_results, xdr_location))) {
 		rstat = TRUE;
 	}
 	(void)xdrrec_endofrecord(xdrs, TRUE);
@@ -788,47 +818,17 @@ __rpc_get_local_uid(SVCXPRT *transp, uid_t *uid) {
  * rpcbind are known to call this function.  Do not alter or remove this
  * API without changing the library's sonum.
  */
+/* Since this is an exported interface used by rpcbind, we cannot
+   remove it. But since poll() can handle much more and much higher
+   file descriptors, this code doesn't really work anymore, too.
+   So for now, keep it as dummy function and do nothing to not break
+   existing binaries. If we have ported rpcbind to the poll() interface
+   and find out, that we really need this cleanup stuff (but nobody
+   besides FreeBSD has this), we need to re-implement it using poll().
+   But this means a new function name with different parameters. For
+   ABI/API compatibility, we cannot reuse this one. */
 bool_t
 __svc_clean_idle(fd_set *fds, int timeout, bool_t cleanblock)
 {
-	int i, ncleaned;
-	SVCXPRT *xprt, *least_active;
-	struct timeval tv, tdiff, tmax;
-	struct cf_conn *cd;
-
-	gettimeofday(&tv, NULL);
-	tmax.tv_sec = tmax.tv_usec = 0;
-	least_active = NULL;
-	rwlock_wrlock(&svc_fd_lock);
-	for (i = ncleaned = 0; i <= svc_maxfd; i++) {
-		if (FD_ISSET(i, fds)) {
-			xprt = __svc_xports[i];
-			if (xprt == NULL || xprt->xp_ops == NULL ||
-			    xprt->xp_ops->xp_recv != svc_vc_recv)
-				continue;
-			cd = (struct cf_conn *)xprt->xp_p1;
-			if (!cleanblock && !cd->nonblock)
-				continue;
-			if (timeout == 0) {
-				timersub(&tv, &cd->last_recv_time, &tdiff);
-				if (timercmp(&tdiff, &tmax, >)) {
-					tmax = tdiff;
-					least_active = xprt;
-				}
-				continue;
-			}
-			if (tv.tv_sec - cd->last_recv_time.tv_sec > timeout) {
-				__xprt_unregister_unlocked(xprt);
-				__svc_vc_dodestroy(xprt);
-				ncleaned++;
-			}
-		}
-	}
-	if (timeout == 0 && least_active != NULL) {
-		__xprt_unregister_unlocked(least_active);
-		__svc_vc_dodestroy(least_active);
-		ncleaned++;
-	}
-	rwlock_unlock(&svc_fd_lock);
-	return ncleaned > 0 ? TRUE : FALSE;
+	return FALSE;
 }
